import typing
import discord
from ..command import Command
from ..commandinterface import CommandInterface
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ...utils import NetworkHandler
from ...network import Request, ResponseSuccess
from ...error import NoneFoundError
if typing.TYPE_CHECKING:
    from ...bots import DiscordBot


class PauseNH(NetworkHandler):
    message_type = "music_pause"

    # noinspection PyProtectedMember
    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        # Find the matching guild
        if data["guild_name"]:
            guild = bot.client.find_guild_by_name(data["guild_name"])
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Set the currently playing source as ended
        voice_client: discord.VoiceClient = bot.client.find_voice_client_by_guild(guild)
        if not (voice_client.is_playing() or voice_client.is_paused()):
            raise NoneFoundError("Nothing to pause")
        # Toggle pause
        resume = voice_client._player.is_paused()
        if resume:
            voice_client._player.resume()
        else:
            voice_client._player.pause()
        return ResponseSuccess({"resumed": resume})


class PauseCommand(Command):
    name: str = "pause"

    description: str = "Mette in pausa o riprende la riproduzione della canzone attuale."

    syntax = "[ [guild] ]"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        interface.register_net_handler(PauseNH.message_type, PauseNH)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        guild, = args.match(r"(?:\[(.+)])?")
        response = await self.interface.net_request(Request("music_pause", {"guild_name": guild}), "discord")
        if response["resumed"]:
            await data.reply(f"▶️ Riproduzione ripresa.")
        else:
            await data.reply(f"⏸ Riproduzione messa in pausa.")



"""A Royal Games Wiki viewer :py:class:`royalnet.web.Royalprint`. Doesn't support any kind of edit."""

import flask as f
import os
from ...royalprint import Royalprint
from ...shortcuts import error, from_urluuid
from ....database.tables import Royal, WikiPage, WikiRevision
from ....utils.wikirender import prepare_page_markdown, RenderError


tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
rp = Royalprint("wikiview", __name__, url_prefix="/wiki/view", template_folder=tmpl_dir,
                required_tables={Royal, WikiPage, WikiRevision})


def prepare_page(page):
    try:
        if page.format == "markdown":
            return f.render_template("wikiview_page.html",
                                     page=page,
                                     parsed_contents=f.Markup(prepare_page_markdown(page.contents)),
                                     css=page.css)
        elif page.format == "html":
            return f.render_template("wikiview_page.html",
                                     page=page,
                                     parsed_contents=f.Markup(page.contents),
                                     css=page.css)
        else:
            return error(500, f"Non esiste nessun handler in grado di preparare pagine con il formato {page.format}.")
    except RenderError as e:
        return error(500, f"La pagina Wiki non può essere preparata a causa di un errore: {str(e)}")


@rp.route("/")
def wikiview_index():
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    pages = sorted(alchemy_session.query(alchemy.WikiPage).all(), key=lambda page: page.title)
    return f.render_template("wikiview_index.html", pages=pages)


@rp.route("/<page_id>", defaults={"title": ""})
@rp.route("/<page_id>/<title>")
def wikiview_by_id(page_id: str, title: str):
    page_uuid = from_urluuid(page_id)
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    page = alchemy_session.query(alchemy.WikiPage).filter(alchemy.WikiPage.page_id == page_uuid).one_or_none()
    if page is None:
        return error(404, f"La pagina richiesta non esiste.")
    return prepare_page(page)



import typing
from ..command import Command
from ..commandargs import CommandArgs
from ..commanddata import CommandData


class PingCommand(Command):
    name: str = "ping"

    description: str = "Gioca a ping-pong con il bot."

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        await data.reply("🏓 Pong!")



from .commandinterface import CommandInterface
from .command import Command
from .commanddata import CommandData
from .commandargs import CommandArgs

__all__ = ["CommandInterface", "Command", "CommandData", "CommandArgs"]



"""A Royal Games Wiki viewer :py:class:`royalnet.web.Royalprint`. Doesn't support any kind of edit."""
import flask as f
import uuid
import os
import datetime
import difflib
from ...royalprint import Royalprint
from ...shortcuts import error, from_urluuid
from ....database.tables import Royal, WikiPage, WikiRevision


tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
rp = Royalprint("wikiedit", __name__, url_prefix="/wiki/edit", template_folder=tmpl_dir,
                required_tables={Royal, WikiPage, WikiRevision})


@rp.route("/newpage", methods=["GET", "POST"])
def wikiedit_newpage():
    if "royal" not in f.session:
        return error(403, "Devi aver effettuato il login per creare pagine wiki.")

    if f.request.method == "GET":
        return f.render_template("wikiedit_page.html", page=None)

    elif f.request.method == "POST":
        fd = f.request.form
        if not ("title" in fd and "contents" in fd and "css" in fd):
            return error(400, "Uno dei campi obbligatori non è stato compilato. Controlla e riprova!")
        alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
        page = alchemy.WikiPage(page_id=uuid.uuid4(),
                                title=fd["title"],
                                contents=fd["contents"],
                                format="markdown",
                                css=fd["css"] if fd["css"] != "None" else None)
        revision = alchemy.WikiRevision(revision_id=uuid.uuid4(),
                                        page=page,
                                        author_id=f.session["royal"]["uid"],
                                        timestamp=datetime.datetime.now(),
                                        reason=fd.get("reason"),
                                        diff="\n".join(difflib.unified_diff([], page.contents.split("\n"))))
        alchemy_session.add(page)
        alchemy_session.add(revision)
        alchemy_session.commit()
        return f.redirect(f.url_for("wikiview.wikiview_by_id", page_id=page.page_short_id, title=page.title))


@rp.route("/<page_id>", defaults={"title": ""}, methods=["GET", "POST"])
@rp.route("/<page_id>/<title>", methods=["GET", "POST"])
def wikiedit_by_id(page_id: str, title: str):
    page_uuid = from_urluuid(page_id)
    if "royal" not in f.session:
        return error(403, "Devi aver effettuato il login per modificare pagine wiki.")

    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    page = alchemy_session.query(alchemy.WikiPage).filter(alchemy.WikiPage.page_id == page_uuid).one_or_none()
    if page is None:
        return error(404, "La pagina che stai cercando di modificare non esiste.")

    if f.request.method == "GET":
        return f.render_template("wikiedit_page.html", page=page)

    elif f.request.method == "POST":
        fd = f.request.form
        if not ("title" in fd and "contents" in fd and "css" in fd):
            return error(400, "Uno dei campi obbligatori non è stato compilato. Controlla e riprova!")
        # Create new revision
        revision = alchemy.WikiRevision(revision_id=uuid.uuid4(),
                                        page=page,
                                        author_id=f.session["royal"]["uid"],
                                        timestamp=datetime.datetime.now(),
                                        reason=fd.get("reason"),
                                        diff="\n".join(difflib.unified_diff(page.contents.split("\n"), fd["contents"].split("\n"))))
        alchemy_session.add(revision)
        # Apply changes
        page.contents = fd["contents"]
        page.title = fd["title"]
        page.css = fd["css"] if fd["css"] != "None" else None
        alchemy_session.commit()
        return f.redirect(f.url_for("wikiview.wikiview_by_id", page_id=page.page_short_id, title=page.title))



import typing
import pickle
from ..command import Command
from ..commandinterface import CommandInterface
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ...utils import NetworkHandler, asyncify
from ...network import Request, ResponseSuccess
from ...error import *
from ...audio import YtdlDiscord
if typing.TYPE_CHECKING:
    from ...bots import DiscordBot


class PlayNH(NetworkHandler):
    message_type = "music_play"

    ytdl_args = {
        "format": "bestaudio",
        "outtmpl": f"./downloads/%(title)s.%(ext)s"
    }

    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        """Handle a play Royalnet request. That is, add audio to a PlayMode."""
        # Find the matching guild
        if data["guild_name"]:
            guild = bot.client.find_guild(data["guild_name"])
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Ensure the guild has a PlayMode before adding the file to it
        if not bot.music_data.get(guild):
            # TODO: change Exception
            raise Exception("No music_data for this guild")
        # Start downloading
        if data["url"].startswith("http://") or data["url"].startswith("https://"):
            dfiles: typing.List[YtdlDiscord] = await asyncify(YtdlDiscord.create_and_ready_from_url, data["url"], **cls.ytdl_args)
        else:
            dfiles = await asyncify(YtdlDiscord.create_and_ready_from_url, f"ytsearch:{data['url']}", **cls.ytdl_args)
        await bot.add_to_music_data(dfiles, guild)
        # Create response dictionary
        response = {
            "videos": [{
                "title": dfile.info.title,
                "discord_embed_pickle": str(pickle.dumps(dfile.info.to_discord_embed()))
            } for dfile in dfiles]
        }
        return ResponseSuccess(response)


class PlayCommand(Command):
    name: str = "play"

    description: str = "Aggiunge una canzone alla coda della chat vocale."

    syntax = "[ [guild] ] (url)"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        interface.register_net_handler(PlayNH.message_type, PlayNH)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        guild_name, url = args.match(r"(?:\[(.+)])?\s*<?(.+)>?")
        response = await self.interface.net_request(Request("music_play", {"url": url, "guild_name": guild_name}), "discord")
        for video in response["videos"]:
            if self.interface.name == "discord":
                # This is one of the unsafest things ever
                embed = pickle.loads(eval(video["discord_embed_pickle"]))
                await data.message.channel.send(content="▶️ Aggiunto alla coda:", embed=embed)
            else:
                await data.reply(f"▶️ Aggiunto alla coda: [i]{video['title']}[/i]")



import discord
import typing
from ..command import Command
from ..commandinterface import CommandInterface
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ...network import Request, ResponseSuccess
from ...utils import NetworkHandler, andformat
from ...bots import DiscordBot
from ...error import *


class CvNH(NetworkHandler):
    message_type = "discord_cv"

    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        # Find the matching guild
        if data["guild_name"]:
            guild: discord.Guild = bot.client.find_guild_by_name(data["guild_name"])
        else:
            if len(bot.client.guilds) == 0:
                raise NoneFoundError("No guilds found")
            if len(bot.client.guilds) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.client.guilds)[0]
        # Edit the message, sorted by channel
        discord_members = list(guild.members)
        channels = {0: None}
        members_in_channels = {0: []}
        message = ""
        # Find all the channels
        for member in discord_members:
            if member.voice is not None:
                channel = members_in_channels.get(member.voice.channel.id)
                if channel is None:
                    members_in_channels[member.voice.channel.id] = list()
                    channel = members_in_channels[member.voice.channel.id]
                    channels[member.voice.channel.id] = member.voice.channel
                channel.append(member)
            else:
                members_in_channels[0].append(member)
        # Edit the message, sorted by channel
        for channel in sorted(channels, key=lambda c: -c):
            members_in_channels[channel].sort(key=lambda x: x.nick if x.nick is not None else x.name)
            if channel == 0 and len(members_in_channels[0]) > 0:
                message += "[b]Non in chat vocale:[/b]\n"
            else:
                message += f"[b]In #{channels[channel].name}:[/b]\n"
            for member in members_in_channels[channel]:
                member: typing.Union[discord.User, discord.Member]
                # Ignore not-connected non-notable members
                if not data["everyone"] and channel == 0 and len(member.roles) < 2:
                    continue
                # Ignore offline members
                if member.status == discord.Status.offline and member.voice is None:
                    continue
                # Online status emoji
                if member.bot:
                    message += "🤖 "
                elif member.status == discord.Status.online:
                    message += "🔵 "
                elif member.status == discord.Status.idle:
                    message += "⚫️ "
                elif member.status == discord.Status.dnd:
                    message += "🔴 "
                elif member.status == discord.Status.offline:
                    message += "⚪️ "
                # Voice
                if channel != 0:
                    # Voice status
                    if member.voice.afk:
                        message += "💤 "
                    elif member.voice.self_deaf or member.voice.deaf:
                        message += "🔇 "
                    elif member.voice.self_mute or member.voice.mute:
                        message += "🔈 "
                    elif member.voice.self_video:
                        message += "📺 "
                    else:
                        message += "🔊 "
                # Nickname
                if member.nick is not None:
                    message += f"[i]{member.nick}[/i]"
                else:
                    message += member.name
                # Game or stream
                if member.activity is not None:
                    if member.activity.type == discord.ActivityType.playing:
                        message += f" | 🎮 {member.activity.name}"
                        # Rich presence
                        try:
                            if member.activity.state is not None:
                                message += f" ({member.activity.state}" \
                                           f" | {member.activity.details})"
                        except AttributeError:
                            pass
                    elif member.activity.type == discord.ActivityType.streaming:
                        message += f" | 📡 {member.activity.url}"
                    elif member.activity.type == discord.ActivityType.listening:
                        if isinstance(member.activity, discord.Spotify):
                            if member.activity.title == member.activity.album:
                                message += f" | 🎧 {member.activity.title} ({andformat(member.activity.artists, final=' e ')})"
                            else:
                                message += f" | 🎧 {member.activity.title} ({member.activity.album} | {andformat(member.activity.artists, final=' e ')})"
                        else:
                            message += f" | 🎧 {member.activity.name}"
                    elif member.activity.type == discord.ActivityType.watching:
                        message += f" | 📺 {member.activity.name}"
                    else:
                        message += f" | ❓ Unknown activity"
                message += "\n"
            message += "\n"
        return ResponseSuccess({"response": message})


class CvCommand(Command):
    name: str = "cv"

    description: str = "Elenca le persone attualmente connesse alla chat vocale."

    syntax: str = "[guildname] "

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        interface.register_net_handler("discord_cv", CvNH)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        # noinspection PyTypeChecker
        guild_name, everyone = args.match(r"(?:\[(.+)])?\s*(\S+)?\s*")
        response = await self.interface.net_request(Request("discord_cv", {"guild_name": guild_name,
                                                                           "everyone": bool(everyone)}),
                                                    destination="discord")
        await data.reply(response["response"])



import typing
import re
import datetime
import telegram
import os
import aiohttp
from ..command import Command
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ...database.tables import Royal, Diario, Alias
from ...utils import asyncify
from ...error import *


async def to_imgur(photosizes: typing.List[telegram.PhotoSize], caption="") -> str:
    # Select the largest photo
    largest_photo = sorted(photosizes, key=lambda p: p.width * p.height)[-1]
    # Get the photo url
    photo_file: telegram.File = await asyncify(largest_photo.get_file)
    # Forward the url to imgur, as an upload
    try:
        imgur_api_key = os.environ["IMGUR_CLIENT_ID"]
    except KeyError:
        raise InvalidConfigError("Missing IMGUR_CLIENT_ID envvar, can't upload images to imgur.")
    async with aiohttp.request("post", "https://api.imgur.com/3/upload", data={
        "image": photo_file.file_path,
        "type": "URL",
        "title": "Diario image",
        "description": caption
    }, headers={
        "Authorization": f"Client-ID {imgur_api_key}"
    }) as request:
        response = await request.json()
        if not response["success"]:
            raise ExternalError("imgur returned an error in the image upload.")
        return response["data"]["link"]


class DiarioCommand(Command):
    name: str = "diario"

    description: str = "Aggiungi una citazione al Diario."

    syntax = "[!] \"(testo)\" --[autore], [contesto]"

    require_alchemy_tables = {Royal, Diario, Alias}

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        if self.interface.name == "telegram":
            update: telegram.Update = data.update
            message: telegram.Message = update.message
            reply: telegram.Message = message.reply_to_message
            creator = await data.get_author()
            # noinspection PyUnusedLocal
            quoted: typing.Optional[str]
            # noinspection PyUnusedLocal
            text: typing.Optional[str]
            # noinspection PyUnusedLocal
            context: typing.Optional[str]
            # noinspection PyUnusedLocal
            timestamp: datetime.datetime
            # noinspection PyUnusedLocal
            media_url: typing.Optional[str]
            # noinspection PyUnusedLocal
            spoiler: bool
            if creator is None:
                await data.reply("⚠️ Devi essere registrato a Royalnet per usare questo comando!")
                return
            if reply is not None:
                # Get the message text
                text = reply.text
                # Check if there's an image associated with the reply
                photosizes: typing.Optional[typing.List[telegram.PhotoSize]] = reply.photo
                if photosizes:
                    # Text is a caption
                    text = reply.caption
                    media_url = await to_imgur(photosizes, text if text is not None else "")
                else:
                    media_url = None
                # Ensure there is a text or an image
                if not (text or media_url):
                    raise InvalidInputError("Missing text.")
                # Find the Royalnet account associated with the sender
                quoted_tg = await asyncify(self.interface.session.query(self.interface.alchemy.Telegram)
                                           .filter_by(tg_id=reply.from_user.id)
                                           .one_or_none)
                quoted_account = quoted_tg.royal if quoted_tg is not None else None
                # Find the quoted name to assign
                quoted_user: telegram.User = reply.from_user
                quoted = quoted_user.full_name
                # Get the timestamp
                timestamp = reply.date
                # Set the other properties
                spoiler = False
                context = None
            else:
                # Get the current timestamp
                timestamp = datetime.datetime.now()
                # Get the message text
                raw_text = " ".join(args)
                # Check if there's an image associated with the reply
                photosizes: typing.Optional[typing.List[telegram.PhotoSize]] = message.photo
                if photosizes:
                    media_url = await to_imgur(photosizes, raw_text if raw_text is not None else "")
                else:
                    media_url = None
                # Parse the text, if it exists
                if raw_text:
                    # Pass the sentence through the diario regex
                    match = re.match(
                        r'(!)? *["«‘“‛‟❛❝〝＂`]([^"]+)["»’”❜❞〞＂`] *(?:(?:-{1,2}|—) *([^,]+))?(?:, *([^ ].*))?',
                        raw_text)
                    # Find the corresponding matches
                    if match is not None:
                        spoiler = bool(match.group(1))
                        text = match.group(2)
                        quoted = match.group(3)
                        context = match.group(4)
                    # Otherwise, consider everything part of the text
                    else:
                        spoiler = False
                        text = raw_text
                        quoted = None
                        context = None
                    # Ensure there's a quoted
                    if not quoted:
                        quoted = None
                    if not context:
                        context = None
                    # Find if there's a Royalnet account associated with the quoted name
                    if quoted is not None:
                        quoted_alias = await asyncify(
                            self.interface.session.query(self.interface.alchemy.Alias)
                                                  .filter_by(alias=quoted.lower()).one_or_none)
                    else:
                        quoted_alias = None
                    quoted_account = quoted_alias.royal if quoted_alias is not None else None
                else:
                    text = None
                    quoted = None
                    quoted_account = None
                    spoiler = False
                    context = None
                # Ensure there is a text or an image
                if not (text or media_url):
                    raise InvalidInputError("Missing text.")
            # Create the diario quote
            diario = self.interface.alchemy.Diario(creator=creator,
                                                   quoted_account=quoted_account,
                                                   quoted=quoted,
                                                   text=text,
                                                   context=context,
                                                   timestamp=timestamp,
                                                   media_url=media_url,
                                                   spoiler=spoiler)
            self.interface.session.add(diario)
            await asyncify(self.interface.session.commit)
            await data.reply(f"✅ {str(diario)}")
        else:
            # Find the creator of the quotes
            creator = await data.get_author(error_if_none=True)
            # Recreate the full sentence
            raw_text = " ".join(args)
            # Pass the sentence through the diario regex
            match = re.match(r'(!)? *["«‘“‛‟❛❝〝＂`]([^"]+)["»’”❜❞〞＂`] *(?:(?:-{1,2}|—) *([^,]+))?(?:, *([^ ].*))?',
                             raw_text)
            # Find the corresponding matches
            if match is not None:
                spoiler = bool(match.group(1))
                text = match.group(2)
                quoted = match.group(3)
                context = match.group(4)
            # Otherwise, consider everything part of the text
            else:
                spoiler = False
                text = raw_text
                quoted = None
                context = None
            timestamp = datetime.datetime.now()
            # Ensure there is some text
            if not text:
                raise InvalidInputError("Missing text.")
            # Or a quoted
            if not quoted:
                quoted = None
            if not context:
                context = None
            # Find if there's a Royalnet account associated with the quoted name
            if quoted is not None:
                quoted_alias = await asyncify(
                    self.interface.session.query(self.interface.alchemy.Alias)
                                          .filter_by(alias=quoted.lower())
                                          .one_or_none)
            else:
                quoted_alias = None
            quoted_account = quoted_alias.royal if quoted_alias is not None else None
            if quoted_alias is not None and quoted_account is None:
                await data.reply("⚠️ Il nome dell'autore è ambiguo, quindi la riga non è stata aggiunta.\n"
                                 "Per piacere, ripeti il comando con un nome più specifico!")
                return
            # Create the diario quote
            diario = self.interface.alchemy.Diario(creator=creator,
                                                   quoted_account=quoted_account,
                                                   quoted=quoted,
                                                   text=text,
                                                   context=context,
                                                   timestamp=timestamp,
                                                   media_url=None,
                                                   spoiler=spoiler)
            self.interface.session.add(diario)
            await asyncify(self.interface.session.commit)
            await data.reply(f"✅ {str(diario)}")



import sys
import typing
import asyncio
import logging
import sentry_sdk
from ..utils import *
from ..network import *
from ..database import *
from ..commands import *
from ..error import *


log = logging.getLogger(__name__)


class GenericBot:
    """A generic bot class, to be used as base for the other more specific classes, such as
    :py:class:`royalnet.bots.TelegramBot` and :py:class:`royalnet.bots.DiscordBot`. """
    interface_name = NotImplemented

    def _init_commands(self, commands: typing.List[typing.Type[Command]]) -> None:
        """Generate the ``commands`` dictionary required to handle incoming messages, and the ``network_handlers``
        dictionary required to handle incoming requests. """
        log.debug(f"Now binding commands")
        self._Interface = self._interface_factory()
        self._Data = self._data_factory()
        self.commands = {}
        self.network_handlers: typing.Dict[str, typing.Type[NetworkHandler]] = {}
        for SelectedCommand in commands:
            log.debug(f"Binding {SelectedCommand.name}...")
            interface = self._Interface()
            self.commands[f"{interface.prefix}{SelectedCommand.name}"] = SelectedCommand(interface)
        log.debug(f"Successfully bound commands")

    def _interface_factory(self) -> typing.Type[CommandInterface]:
        # noinspection PyAbstractClass,PyMethodParameters
        class GenericInterface(CommandInterface):
            alchemy = self.alchemy
            bot = self
            loop = self.loop

            def register_net_handler(ci, message_type: str, network_handler: typing.Callable):
                self.network_handlers[message_type] = network_handler

            def unregister_net_handler(ci, message_type: str):
                del self.network_handlers[message_type]

            async def net_request(ci, request: Request, destination: str) -> dict:
                if self.network is None:
                    raise InvalidConfigError("Royalnet is not enabled on this bot")
                response_dict: dict = await self.network.request(request.to_dict(), destination)
                if "type" not in response_dict:
                    raise RoyalnetResponseError("Response is missing a type")
                elif response_dict["type"] == "ResponseSuccess":
                    response: typing.Union[ResponseSuccess, ResponseError] = ResponseSuccess.from_dict(response_dict)
                elif response_dict["type"] == "ResponseError":
                    response = ResponseError.from_dict(response_dict)
                else:
                    raise RoyalnetResponseError("Response type is unknown")
                response.raise_on_error()
                return response.data

        return GenericInterface

    def _data_factory(self) -> typing.Type[CommandData]:
        raise NotImplementedError()

    def _init_royalnet(self, royalnet_config: RoyalnetConfig):
        """Create a :py:class:`royalnet.network.RoyalnetLink`, and run it as a :py:class:`asyncio.Task`."""
        self.network: RoyalnetLink = RoyalnetLink(royalnet_config.master_uri, royalnet_config.master_secret,
                                                  self.interface_name, self._network_handler)
        log.debug(f"Running RoyalnetLink {self.network}")
        self.loop.create_task(self.network.run())

    async def _network_handler(self, request_dict: dict) -> dict:
        """Handle a single :py:class:`dict` received from the :py:class:`royalnet.network.RoyalnetLink`.

        Returns:
            Another :py:class:`dict`, formatted as a :py:class:`royalnet.network.Response`."""
        # Convert the dict to a Request
        try:
            request: Request = Request.from_dict(request_dict)
        except TypeError:
            log.warning(f"Invalid request received: {request_dict}")
            return ResponseError("invalid_request",
                                 f"The Request that you sent was invalid. Check extra_info to see what you sent.",
                                 extra_info={"you_sent": request_dict}).to_dict()
        log.debug(f"Received {request} from the RoyalnetLink")
        try:
            network_handler = self.network_handlers[request.handler]
        except KeyError:
            log.warning(f"Missing network_handler for {request.handler}")
            return ResponseError("no_handler", f"This Link is missing a network handler for {request.handler}.").to_dict()
        try:
            log.debug(f"Using {network_handler} as handler for {request.handler}")
            response: Response = await getattr(network_handler, self.interface_name)(self, request.data)
            return response.to_dict()
        except Exception:
            _, exc, _ = sys.exc_info()
            log.debug(f"Exception {exc} in {network_handler}")
            return ResponseError("exception_in_handler",
                                 f"An exception was raised in {network_handler} for {request.handler}. Check "
                                 f"extra_info for details.",
                                 extra_info={
                                     "type": exc.__class__.__name__,
                                     "str": str(exc)
                                 }).to_dict()

    def _init_database(self, commands: typing.List[typing.Type[Command]], database_config: DatabaseConfig):
        """Create an :py:class:`royalnet.database.Alchemy` with the tables required by the commands. Then,
        find the chain that links the ``master_table`` to the ``identity_table``. """
        log.debug(f"Initializing database")
        required_tables = {database_config.master_table, database_config.identity_table}
        for command in commands:
            required_tables = required_tables.union(command.require_alchemy_tables)
        log.debug(f"Found {len(required_tables)} required tables")
        self.alchemy = Alchemy(database_config.database_uri, required_tables)
        self.master_table = self.alchemy.__getattribute__(database_config.master_table.__name__)
        self.identity_table = self.alchemy.__getattribute__(database_config.identity_table.__name__)
        self.identity_column = self.identity_table.__getattribute__(self.identity_table,
                                                                    database_config.identity_column_name)
        self.identity_chain = relationshiplinkchain(self.master_table, self.identity_table)
        log.debug(f"Identity chain is {self.identity_chain}")

    def __init__(self, *,
                 royalnet_config: typing.Optional[RoyalnetConfig] = None,
                 database_config: typing.Optional[DatabaseConfig] = None,
                 commands: typing.List[typing.Type[Command]] = None,
                 sentry_dsn: typing.Optional[str] = None,
                 loop: asyncio.AbstractEventLoop = None):
        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop
        if sentry_dsn:
            log.debug("Sentry integration enabled")
            self.sentry = sentry_sdk.init(sentry_dsn)
        else:
            log.debug("Sentry integration disabled")
        try:
            if database_config is None:
                self.alchemy = None
                self.master_table = None
                self.identity_table = None
                self.identity_column = None
            else:
                self._init_database(commands=commands, database_config=database_config)
            if commands is None:
                commands = []
            self._init_commands(commands)
            if royalnet_config is None:
                self.network = None
            else:
                self._init_royalnet(royalnet_config=royalnet_config)
        except Exception as e:
            sentry_sdk.capture_exception(e)

    async def run(self):
        """A blocking coroutine that should make the bot start listening to commands and requests."""
        raise NotImplementedError()



"""The production Royalnet, active at @royalgamesbot on Telegram and Royalbot on Discord."""

import os
import asyncio
import logging
import sentry_sdk
from royalnet.bots import DiscordBot, DiscordConfig, TelegramBot, TelegramConfig
from royalnet.commands.royalgames import *
from royalnet.commands.debug import *
from royalnet.network import RoyalnetServer, RoyalnetConfig
from royalnet.database import DatabaseConfig
from royalnet.database.tables import Royal, Telegram, Discord

loop = asyncio.get_event_loop()

log = logging.getLogger("royalnet.bots")
stream_handler = logging.StreamHandler()
stream_handler.formatter = logging.Formatter("{asctime}\t{name}\t{levelname}\t{message}", style="{")
log.addHandler(stream_handler)


sentry_dsn = os.environ.get("SENTRY_DSN")

# noinspection PyUnreachableCode
if __debug__:
    commands = [

    ]
    log.setLevel(logging.DEBUG)
else:
    commands = [
        CiaoruoziCommand,
        ColorCommand,
        CvCommand,
        DiarioCommand,
        Mp3Command,
        PauseCommand,
        PingCommand,
        PlayCommand,
        PlaymodeCommand,
        QueueCommand,
        RageCommand,
        ReminderCommand,
        ShipCommand,
        SkipCommand,
        SmecdsCommand,
        SummonCommand,
        VideochannelCommand,
        DnditemCommand,
        DndspellCommand,
        TriviaCommand
    ]
    log.setLevel(logging.INFO)

address, port = "127.0.0.1", 1234

print("Starting master...")
master = RoyalnetServer(address, port, os.environ["MASTER_KEY"])
loop.run_until_complete(master.start())

print("Starting bots...")
ds_bot = DiscordBot(discord_config=DiscordConfig(os.environ["DS_AK"]),
                    royalnet_config=RoyalnetConfig(f"ws://{address}:{port}", os.environ["MASTER_KEY"]),
                    database_config=DatabaseConfig(os.environ["DB_PATH"], Royal, Discord, "discord_id"),
                    sentry_dsn=sentry_dsn,
                    commands=commands)
tg_bot = TelegramBot(telegram_config=TelegramConfig(os.environ["TG_AK"]),
                     royalnet_config=RoyalnetConfig(f"ws://{address}:{port}", os.environ["MASTER_KEY"]),
                     database_config=DatabaseConfig(os.environ["DB_PATH"], Royal, Telegram, "tg_id"),
                     sentry_dsn=sentry_dsn,
                     commands=commands)
loop.create_task(tg_bot.run())
loop.create_task(ds_bot.run())

print("Enabled commands:")
for command in commands:
    print(f"{command.name} - {command.description}")

print("Running loop...")
loop.run_forever()



import typing
import pickle
import discord
from ..command import Command
from ..commandinterface import CommandInterface
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ...utils import NetworkHandler, asyncify
from ...network import Request, ResponseSuccess
from ...error import *
from ...audio import YtdlDiscord
if typing.TYPE_CHECKING:
    from ...bots import DiscordBot


class SkipNH(NetworkHandler):
    message_type = "music_skip"

    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        # Find the matching guild
        if data["guild_name"]:
            guild = bot.client.find_guild_by_name(data["guild_name"])
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Set the currently playing source as ended
        voice_client: discord.VoiceClient = bot.client.find_voice_client_by_guild(guild)
        if not (voice_client.is_playing() or voice_client.is_paused()):
            raise NoneFoundError("Nothing to skip")
        # noinspection PyProtectedMember
        voice_client._player.stop()
        return ResponseSuccess()


class SkipCommand(Command):
    name: str = "skip"

    description: str = "Salta la canzone attualmente in riproduzione in chat vocale."

    syntax: str = "[ [guild] ]"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        interface.register_net_handler(SkipNH.message_type, SkipNH)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        guild, = args.match(r"(?:\[(.+)])?")
        await self.interface.net_request(Request(SkipNH.message_type, {"guild_name": guild}), "discord")
        await data.reply(f"⏩ Richiesto lo skip della canzone attuale.")



import telegram
import telegram.utils.request
import typing
import uuid
import asyncio
import sentry_sdk
import logging as _logging
from .generic import GenericBot
from ..utils import *
from ..error import *
from ..network import *
from ..database import *
from ..commands import *


log = _logging.getLogger(__name__)


class TelegramConfig:
    """The specific configuration to be used for :py:class:`royalnet.database.TelegramBot`."""
    def __init__(self, token: str):
        self.token: str = token


class TelegramBot(GenericBot):
    """A bot that connects to `Telegram <https://telegram.org/>`_."""
    interface_name = "telegram"

    def _init_client(self):
        """Create the :py:class:`telegram.Bot`, and set the starting offset."""
        # https://github.com/python-telegram-bot/python-telegram-bot/issues/341
        request = telegram.utils.request.Request(5)
        self.client = telegram.Bot(self._telegram_config.token, request=request)
        self._offset: int = -100

    def _interface_factory(self) -> typing.Type[CommandInterface]:
        # noinspection PyPep8Naming
        GenericInterface = super()._interface_factory()

        # noinspection PyMethodParameters,PyAbstractClass
        class TelegramInterface(GenericInterface):
            name = self.interface_name
            prefix = "/"

            def __init__(self):
                super().__init__()
                self.keys_callbacks: typing.Dict[typing.Tuple[int, str], typing.Callable] = {}

            def register_keyboard_key(interface, key_name: str, callback: typing.Callable):
                interface.keys_callbacks[key_name] = callback

            def unregister_keyboard_key(interface, key_name: str):
                del interface.keys_callbacks[key_name]

        return TelegramInterface

    def _data_factory(self) -> typing.Type[CommandData]:
        # noinspection PyMethodParameters,PyAbstractClass
        class TelegramData(CommandData):
            def __init__(data, interface: CommandInterface, update: telegram.Update):
                data.interface = interface
                data.update = update

            async def reply(data, text: str):
                await asyncify(data.update.effective_chat.send_message, telegram_escape(text),
                               parse_mode="HTML",
                               disable_web_page_preview=True)

            async def get_author(data, error_if_none=False):
                user: telegram.User = data.update.effective_user
                if user is None:
                    if error_if_none:
                        raise UnregisteredError("No author for this message")
                    return None
                query = data.interface.session.query(self.master_table)
                for link in self.identity_chain:
                    query = query.join(link.mapper.class_)
                query = query.filter(self.identity_column == user.id)
                result = await asyncify(query.one_or_none)
                if result is None and error_if_none:
                    raise UnregisteredError("Author is not registered")
                return result

            async def keyboard(data, text: str, keyboard: typing.Dict[str, typing.Callable]) -> None:
                tg_keyboard = []
                for key in keyboard:
                    press_id = uuid.uuid4()
                    tg_keyboard.append([telegram.InlineKeyboardButton(key, callback_data=str(press_id))])
                    data.interface.register_keyboard_key(key_name=str(press_id), callback=keyboard[key])
                await asyncify(data.update.effective_chat.send_message,
                               telegram_escape(text),
                               reply_markup=telegram.InlineKeyboardMarkup(tg_keyboard),
                               parse_mode="HTML",
                               disable_web_page_preview=True)

        return TelegramData

    def __init__(self, *,
                 telegram_config: TelegramConfig,
                 royalnet_config: typing.Optional[RoyalnetConfig] = None,
                 database_config: typing.Optional[DatabaseConfig] = None,
                 sentry_dsn: typing.Optional[str] = None,
                 commands: typing.List[typing.Type[Command]] = None):
        super().__init__(royalnet_config=royalnet_config,
                         database_config=database_config,
                         sentry_dsn=sentry_dsn,
                         commands=commands)
        self._telegram_config = telegram_config
        self._init_client()

    async def _handle_update(self, update: telegram.Update):
        # Skip non-message updates
        if update.message is not None:
            await self._handle_message(update)
        elif update.callback_query is not None:
            await self._handle_callback_query(update)

    async def _handle_message(self, update: telegram.Update):
        message: telegram.Message = update.message
        text: str = message.text
        # Try getting the caption instead
        if text is None:
            text: str = message.caption
        # No text or caption, ignore the message
        if text is None:
            return
        # Skip non-command updates
        if not text.startswith("/"):
            return
        # Find and clean parameters
        command_text, *parameters = text.split(" ")
        command_name = command_text.replace(f"@{self.client.username}", "").lower()
        # Send a typing notification
        update.message.chat.send_action(telegram.ChatAction.TYPING)
        # Find the command
        try:
            command = self.commands[command_name]
        except KeyError:
            # Skip the message
            return
        # Prepare data
        data = self._Data(interface=command.interface, update=update)
        # Run the command
        try:
            await command.run(CommandArgs(parameters), data)
        except Exception as e:
            sentry_sdk.capture_exception(e)
            error_message = f"⛔️ [b]{e.__class__.__name__}[/b]\n"
            error_message += '\n'.join(e.args)
            await data.reply(error_message)

    async def _handle_callback_query(self, update: telegram.Update):
        query: telegram.CallbackQuery = update.callback_query
        source: telegram.Message = query.message
        callback: typing.Optional[typing.Callable] = None
        command: typing.Optional[Command] = None
        for command in self.commands.values():
            if query.data in command.interface.keys_callbacks:
                callback = command.interface.keys_callbacks[query.data]
                break
        if callback is None:
            await asyncify(source.edit_reply_markup, reply_markup=None)
            await asyncify(query.answer, text="⛔️ This keyboard has expired.")
            return
        try:
            response = await callback(data=self._Data(interface=command.interface, update=update))
        except KeyboardExpiredError:
            # FIXME: May cause a memory leak, as keys are not deleted after use
            await asyncify(source.edit_reply_markup, reply_markup=None)
            await asyncify(query.answer, text="⛔️ This keyboard has expired.")
            return
        except Exception as e:
            error_text = f"⛔️ {e.__class__.__name__}\n"
            error_text += '\n'.join(e.args)
            await asyncify(query.answer, text=error_text)
            return
        else:
            await asyncify(query.answer, text=response)

    async def run(self):
        while True:
            # Get the latest 100 updates
            try:
                last_updates: typing.List[telegram.Update] = await asyncify(self.client.get_updates,
                                                                            offset=self._offset,
                                                                            timeout=30,
                                                                            read_latency=5.0)
            except telegram.error.TimedOut as error:
                log.debug("getUpdates timed out")
                continue
            except Exception as error:
                log.error(f"Error while getting updates: {error.__class__.__name__} {error.args}")
                sentry_sdk.capture_exception(error)
                await asyncio.sleep(5)
                continue
            # Handle updates
            for update in last_updates:
                # noinspection PyAsyncCall
                self.loop.create_task(self._handle_update(update))
            # Recalculate offset
            try:
                self._offset = last_updates[-1].update_id + 1
            except IndexError:
                pass



from . import audio, bots, database, network, utils, error, web, version
from royalnet import commands

__all__ = ["audio", "bots", "commands", "database", "network", "utils", "error", "web", "version"]



import typing


class CommandData:
    async def reply(self, text: str) -> None:
        """Send a text message to the channel where the call was made.

        Parameters:
             text: The text to be sent, possibly formatted in the weird undescribed markup that I'm using."""
        raise NotImplementedError()

    async def get_author(self, error_if_none: bool = False):
        """Try to find the identifier of the user that sent the message.
        That probably means, the database row identifying the user.

        Parameters:
            error_if_none: Raise a :py:exc:`royalnet.error.UnregisteredError` if this is True and the call has no author.

        Raises:
             :py:exc:`royalnet.error.UnregisteredError` if ``error_if_none`` is set to True and no author is found."""
        raise NotImplementedError()

    async def keyboard(self, text: str, keyboard: typing.Dict[str, typing.Callable]) -> None:
        """Send a keyboard having the keys of the dict as keys and calling the correspondent values on a press.

        The function should be passed the :py:class:`CommandData` instance as a argument."""
        raise NotImplementedError()



import typing
import pickle
from ..command import Command
from ..commandinterface import CommandInterface
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ...utils import NetworkHandler, numberemojiformat
from ...network import Request, ResponseSuccess
from ...error import *
if typing.TYPE_CHECKING:
    from ...bots import DiscordBot


class QueueNH(NetworkHandler):
    message_type = "music_queue"

    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        # Find the matching guild
        if data["guild_name"]:
            guild = bot.client.find_guild_by_name(data["guild_name"])
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Check if the guild has a PlayMode
        playmode = bot.music_data.get(guild)
        if not playmode:
            return ResponseSuccess({
                "type": None
            })
        try:
            queue = playmode.queue_preview()
        except NotImplementedError:
            return ResponseSuccess({
                "type": playmode.__class__.__name__
            })
        return ResponseSuccess({
            "type": playmode.__class__.__name__,
            "queue":
            {
                "strings": [str(dfile.info) for dfile in queue],
                "pickled_embeds": str(pickle.dumps([dfile.info.to_discord_embed() for dfile in queue]))
            }
        })


class QueueCommand(Command):
    name: str = "queue"

    description: str = "Visualizza la coda di riproduzione attuale."

    syntax = "[ [guild] ]"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        interface.register_net_handler(QueueNH.message_type, QueueNH)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        guild, = args.match(r"(?:\[(.+)])?")
        response = await self.interface.net_request(Request(QueueNH.message_type, {"guild_name": guild}), "discord")
        if response["type"] is None:
            await data.reply("ℹ️ Non c'è nessuna coda di riproduzione attiva al momento.")
            return
        elif "queue" not in response:
            await data.reply(f"ℹ️ La coda di riproduzione attuale ([c]{response['type']}[/c]) non permette l'anteprima.")
            return
        if response["type"] == "Playlist":
            if len(response["queue"]["strings"]) == 0:
                message = f"ℹ️ Questa [c]Playlist[/c] è vuota."
            else:
                message = f"ℹ️ Questa [c]Playlist[/c] contiene {len(response['queue']['strings'])} elementi, e i prossimi saranno:\n"
        elif response["type"] == "Pool":
            if len(response["queue"]["strings"]) == 0:
                message = f"ℹ️ Questo [c]Pool[/c] è vuoto."
            else:
                message = f"ℹ️ Questo [c]Pool[/c] contiene {len(response['queue']['strings'])} elementi, tra cui:\n"
        elif response["type"] == "Layers":
            if len(response["queue"]["strings"]) == 0:
                message = f"ℹ️ Nessun elemento è attualmente in riproduzione, pertanto non ci sono [c]Layers[/c]:"
            else:
                message = f"ℹ️ I [c]Layers[/c] dell'elemento attualmente in riproduzione sono {len(response['queue']['strings'])}, tra cui:\n"
        else:
            if len(response["queue"]["strings"]) == 0:
                message = f"ℹ️ Il PlayMode attuale, [c]{response['type']}[/c], è vuoto.\n"
            else:
                message = f"ℹ️ Il PlayMode attuale, [c]{response['type']}[/c], contiene {len(response['queue']['strings'])} elementi:\n"
        if self.interface.name == "discord":
            await data.reply(message)
            for embed in pickle.loads(eval(response["queue"]["pickled_embeds"]))[:5]:
                await data.message.channel.send(embed=embed)
        else:
            message += numberemojiformat(response["queue"]["strings"][:10])
            await data.reply(message)



import typing
import aiohttp
import sortedcontainers
from ..command import Command
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ..commandinterface import CommandInterface
from ...utils import parse_5etools_entry


class DnditemCommand(Command):
    name: str = "dnditem"

    description: str = "Ottieni informazioni su un oggetto di D&D5e."

    syntax = "(nomeoggetto)"

    _dnddata: sortedcontainers.SortedKeyList = None

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        interface.loop.create_task(self._fetch_dnddata())

    async def _fetch_dnddata(self):
        self._dnddata = self._dnddata = sortedcontainers.SortedKeyList([], key=lambda i: i["name"].lower())
        async with aiohttp.ClientSession() as session:
            async with session.get("https://5e.tools/data/items.json") as response:
                j = await response.json()
                for item in j["item"]:
                    self._dnddata.add(item)
            async with session.get("https://5e.tools/data/fluff-items.json") as response:
                j = await response.json()
                for item in j["item"]:
                    self._dnddata.add(item)
            async with session.get("https://5e.tools/data/items-base.json") as response:
                j = await response.json()
                for item in j["baseitem"]:
                    self._dnddata.add(item)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        if self._dnddata is None:
            await data.reply("⚠️ Il database degli oggetti di D&D non è ancora stato scaricato.")
            return
        search = args.joined().lower()
        result = self._dnddata[self._dnddata.bisect_key_left(search)]
        string = f'📦 [b]{result["name"]}[/b]\n'
        if "source" in result:
            string += f'[i]{result["source"]}, page {result["page"]}[/i]\n'
        string += f'\n' \
                  f'Type: [b]{result.get("type", "None")}[/b]\n' \
                  f'Value: [b]{result.get("value", "-")}[/b]\n' \
                  f'Weight: [b]{result.get("weight", "0")} lb[/b]\n' \
                  f'Rarity: [b]{result["rarity"] if result.get("rarity", "None") != "None" else "Mundane"}[/b]\n' \
                  f'\n'
        for entry in result.get("entries", []):
            string += parse_5etools_entry(entry)
            string += "\n\n"
        await data.reply(string)
